# Scola version 1.1.0
# ScolaWizard.py
# by Antoine Rauzy
# Copyrights (c) 2018 AltaRica Association
#
# This file is part of ScolaWizard.
# ScolaWizard is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# ScolaWizard is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with ScolaWizard.  If not, see <http://www.gnu.org/licenses/>.
 
# Table of Contents
# -----------------
# 1) Imported Modules
# 2) Creator
# 3) Callbacks
# 4) Update
# 5) File redirector

# 1) Imported Modules
# -------------------

import string
import sys
import re
import enum
import math
import random
import tkinter 
import tkinter.ttk
import tkinter.scrolledtext
import tkinter.filedialog
import tkinter.messagebox

from ScolaWalker import *
from ScolaText import *

# 2) Creator
# ----------

class ScolaWizard:
  _wizard = None

  def __init__(self):
    self._walker = ScolaWalker()
    self.zoneSize = 4
    self._modelFileName = None
    self._rootWindow = tkinter.Tk()
    self._rootWindow.title("Scola Wizard")
    self._rootWindow.minsize(640, 320)
    self._rootWindow.wm_iconbitmap(bitmap="src/Wizard/AltaRicaWizard.ico")
    self._CreateMenu()
    self._CreateToolBar()
    self._CreateMainPane()
    # self._CreateCanvas(defaultWidth, defaultHeight)
    self._rootWindow.bind("<Control-o>", ScolaWizard.LoadModelKey)
    self._rootWindow.bind("<Control-s>", ScolaWizard.SaveModelKey)
    self._walker.SetOutputFile(self._redirectorToTextWidgetOutput)
    self._walker.SetErrorFile(self._redirectorToTextWidgetError)

  def _CreateMenu(self):
    self.menuBar = tkinter.Menu(self._rootWindow)
    # Menu Model
    self._modelMenu = tkinter.Menu(self.menuBar, tearoff=0)
    self._modelMenu.add_command(label="Load", command=self.LoadModel)
    self._modelMenu.add_command(label="Reset", command=self.ResetModel)
    self._modelMenu.add_separator()
    self._modelMenu.add_command(label="Show source model", command=self.PrintSourceModel)
    self._modelMenu.add_command(label="Show instantiated model", command=self.PrintInstantiatedModel)
    self._modelMenu.add_separator()
    self._modelMenu.add_command(label="Exit", command=self.Quit)
    self.menuBar.add_cascade(label="Model", menu=self._modelMenu)
    # Menu Walk
    self._walkMenu = tkinter.Menu(self.menuBar, tearoff=0)
    self._walkMenu.add_command(label="Start", command=self.StartWalk)
    self._walkMenu.add_command(label="Stop", command=self.StopWalk)
    self._walkMenu.add_separator()
    self._walkMenu.add_command(label="FindOne", command=self.FindOne)
    self._walkMenu.add_command(label="FindNext", command=self.FindNext)
    self._walkMenu.add_separator()
    self._walkMenu.add_command(label="Save history", command=self.SaveHistory)
    self._walkMenu.add_command(label="Play history", command=self.PlayHistory)
    self._walkMenu.add_separator()
    self._walkMenu.add_command(label="Options", command=self.SetOptions)
    self.menuBar.add_cascade(label="Walk", menu=self._walkMenu)
    # self.displayMenu = tkinter.Menu(self.menuBar, tearoff=0)
    # self.displayMenu.add_command(label="Map Display", command=self.CallMapDisplayForm)
    # self.menuBar.add_cascade(label="Display", menu=self.displayMenu)
    # Menu Help
    self._helpMenu = tkinter.Menu(self.menuBar, tearoff=0)
    self._helpMenu.add_command(label="About", command=self.AboutDialog)
    self.menuBar.add_cascade(label="?", menu=self._helpMenu)
    self._rootWindow.config(menu=self.menuBar)

  def _CreateToolBar(self):
    self.toolBar = tkinter.Frame(self._rootWindow)
    # self.variableHistoryNumber = tkinter.StringVar()
    # self.variableHistoryStep = tkinter.StringVar()
    # self.labelHistoryHeader = tkinter.Label(self.toolBar, text="History", width=8)
    # self.labelHistoryNumber = tkinter.Label(self.toolBar, textvariable=self.variableHistoryNumber, width=8)
    # self.labelHistoryStep = tkinter.Label(self.toolBar, textvariable=self.variableHistoryStep, width=8)
    # self.labelHistoryHeader.pack(side="top", fill="x")
    # self.labelHistoryNumber.pack(side="top", fill="x")
    # self.labelHistoryStep.pack(side="top", fill="x")
    # self.variableWindDirection = tkinter.StringVar()
    # self.variableWindForce = tkinter.StringVar()
    # self.labelWindHeader = tkinter.Label(self.toolBar, text="Wind", width=8)
    # self.labelWindDirection = tkinter.Label(self.toolBar, textvariable=self.variableWindDirection, width=8)
    # self.labelWindForce = tkinter.Label(self.toolBar, textvariable=self.variableWindForce, width=8)
    # self.labelWindHeader.pack(side="top", fill="x")
    # self.labelWindDirection.pack(side="top", fill="x")
    # self.labelWindForce.pack(side="top", fill="x")
    # self.buttonStopSimulation = tkinter.Button(self.toolBar, text="stop", bg="salmon", command=self.RaiseStopSimulation)
    self.toolBar.pack(side="left", fill="y")

  def _CreateMainPane(self):
    self._notebook = tkinter.ttk.Notebook()
    # tab output
    self._frameOutput = tkinter.Frame(self._notebook)
    self._notebook.add(self._frameOutput, text="output", sticky=tkinter.N+tkinter.S+tkinter.E+tkinter.W)
    self._textWidgetOutput = ScolaText(self._frameOutput)
    self._redirectorToTextWidgetOutput = ScolaOutputRedirector(self._textWidgetOutput)
    self._redirectorToTextWidgetError = ScolaOutputRedirector(self._textWidgetOutput, True)
    self._textWidgetOutputScrollBar = tkinter.Scrollbar(self._frameOutput, command=self._textWidgetOutput.yview)
    self._textWidgetOutput.configure(yscrollcommand=self._textWidgetOutputScrollBar.set)
    self._textWidgetOutputScrollBar.pack(side=tkinter.RIGHT, fill=tkinter.Y, expand=False)
    self._textWidgetOutput.pack(side=tkinter.LEFT, fill=tkinter.BOTH, expand=True)
    # tab processes
    self._frameProcesses = tkinter.Frame(self._notebook)
    self._notebook.add(self._frameProcesses, text="processes", sticky=tkinter.N+tkinter.S+tkinter.E+tkinter.W)
    self._textWidgetProcesses = ScolaText(self._frameProcesses)
    self._redirectorToTextWidgetProcesses = ScolaOutputRedirector(self._textWidgetProcesses)
    self._textWidgetProcessesScrollBar = tkinter.Scrollbar(self._frameProcesses, command=self._textWidgetProcesses.yview)
    self._textWidgetProcesses.configure(yscrollcommand=self._textWidgetProcessesScrollBar.set)
    self._textWidgetProcessesScrollBar.pack(side=tkinter.RIGHT, fill=tkinter.Y, expand=False)
    self._textWidgetProcesses.pack(side=tkinter.LEFT, fill=tkinter.BOTH, expand=True)
    # tab state
    self._frameSystem = tkinter.Frame(self._notebook)
    self._notebook.add(self._frameSystem, text="system", sticky=tkinter.N+tkinter.S+tkinter.E+tkinter.W)
    self._textWidgetSystem = ScolaText(self._frameSystem)
    self._redirectorToTextWidgetSystem = ScolaOutputRedirector(self._textWidgetSystem)
    self._textWidgetSystemScrollBar = tkinter.Scrollbar(self._frameSystem, command=self._textWidgetSystem.yview)
    self._textWidgetSystem.configure(yscrollcommand=self._textWidgetSystemScrollBar.set)
    self._textWidgetSystemScrollBar.pack(side=tkinter.RIGHT, fill=tkinter.Y, expand=False)
    self._textWidgetSystem.pack(side=tkinter.LEFT, fill=tkinter.BOTH, expand=True)
    # tab history
    self._frameHistory = tkinter.Frame(self._notebook)
    self._notebook.add(self._frameHistory, text="history", sticky=tkinter.N+tkinter.S+tkinter.E+tkinter.W)
    self._textWidgetHistory = ScolaText(self._frameHistory)
    self._redirectorToTextWidgetHistory = ScolaOutputRedirector(self._textWidgetHistory)
    self._textWidgetHistoryScrollBar = tkinter.Scrollbar(self._frameHistory, command=self._textWidgetHistory.yview)
    self._textWidgetHistory.configure(yscrollcommand=self._textWidgetHistoryScrollBar.set)
    self._textWidgetHistoryScrollBar.pack(side=tkinter.RIGHT, fill=tkinter.Y, expand=False)
    self._textWidgetHistory.pack(side=tkinter.LEFT, fill=tkinter.BOTH, expand=True)

    # simulation frame
      # creation
    self._simulationFrame = tkinter.Frame(borderwidth=2, relief=tkinter.RAISED)
    self._nextButton = tkinter.Button(self._simulationFrame, text="next", command=self.MoveCurrentProcessForward)
    self._currentProcessNumber = tkinter.IntVar()
    self._currentProcessNumberCombobox = tkinter.ttk.Combobox(self._simulationFrame, width=4, state = 'readonly', textvariable=self._currentProcessNumber)
    self._currentProcessBranch = tkinter.StringVar()
    self._currentProcessBranchCombobox = tkinter.ttk.Combobox(self._simulationFrame, width=10, state = 'readonly', textvariable=self._currentProcessBranch)
    self._simulationFrameSeparator = tkinter.ttk.Separator(self._simulationFrame, orient=tkinter.VERTICAL)
    self._backButton = tkinter.Button(self._simulationFrame, text="back", command=self.Backtrack)
      # creation
    self._backButton.pack(side=tkinter.RIGHT, expand=False, padx=10, ipadx=10, pady=5)
    self._simulationFrameSeparator.pack(side=tkinter.RIGHT, expand=False, pady=5, fill=tkinter.Y)
    self._currentProcessBranchCombobox.pack(side=tkinter.RIGHT, expand=False, padx=10, ipadx=10, pady=5)
    self._currentProcessNumberCombobox.pack(side=tkinter.RIGHT, expand=False, padx=10, ipadx=10, pady=5)
    self._nextButton.pack(side=tkinter.RIGHT, expand=False, padx=10, ipadx=10, pady=5)
    self._currentProcessNumberCombobox.bind("<<ComboboxSelected>>", ScolaWizard.UpdateCurrentProcessBranch)
    
    # Packing so to have good behavior when resizing, i.e. simulation frame is always visible
    self._simulationFrame.pack(side=tkinter.BOTTOM, fill=tkinter.X, expand=False)
    self._notebook.pack(side=tkinter.BOTTOM, expand=1, fill=tkinter.BOTH)

  @staticmethod
  def GetWizard():
    if ScolaWizard._wizard==None:
      ScolaWizard._wizard = ScolaWizard()
    return ScolaWizard._wizard

  def Launch(self):
    # random.seed(self.seed)
    self._rootWindow.mainloop()

  def RaiseError(self, message):
    tkinter.messagebox.showerror("error", message)
    sys.stderr.write(message)
    sys.stderr.flush()

  def AboutDialog(self):
    tkinter.messagebox.showinfo("about", """
Scola Wizard: step by step execution of Scola models
by Antoine Rauzy
Copyright (c) 2018 AltaRica Association
The current version of Scola Wizard is version 1.0
Scola Wizard is distributed under GNU GPLv3 license
""")

# 3) Callbacks
# ------------

  def Quit(self):
    self._rootWindow.destroy()

  def LoadModel(self):
    fileName = tkinter.filedialog.askopenfilename(filetypes = (("Scola", "*.scola"),("All files", "*.*") ))
    if fileName=='':
      return
    self._modelFileName = fileName
    self._modelFileName = self._modelFileName.strip('"')
    try:
      inputFile = open(self._modelFileName, "r")
    except:
      self.RaiseError('Unable to open file "%s"\n' % self._modelFileName)
      return
    numberOfErrors = self._walker.ParseFile(self._modelFileName, inputFile)
    inputFile.close()
    if numberOfErrors==0:
      self._redirectorToTextWidgetOutput.write("Loading file %s completed\n" % self._modelFileName)
    else:
      self._redirectorToTextWidgetOutput.write("Loading file %s failed: %d errors detected\n" % (self._modelFileName, numberOfErrors))

  def SaveModel(self):
    if self._modelFileName==None:
      self.SaveModelAs()
      return
    # self.ExportMap(self.mapFileName)

  def SaveModelAs(self):
    fileName = tkinter.filedialog.asksaveasfilename(filetypes = (("Scola", "*.scola"),("All files", "*.*") ))
    if fileName=='':
      return
    # self.ExportMap(fileName)
    self._modelFileName = fileName

  @staticmethod
  def LoadModelKey(key):
    wizard = ScolaWizard.GetWizard()
    wizard.LoadModel()

  @staticmethod
  def SaveModelKey(key):
    wizard = ScolaWizard.GetWizard()
    wizard.SaveModel()

  def ResetModel(self):
    self._walker.ResetModel()
    self._modelFileName = None
    self._redirectorToTextWidgetOutput.write("Model reseted\n")
    self._UpdateDisplay()

  def PrintSourceModel(self):
    self._textWidgetOutput.delete('1.0', tkinter.END)
    self._walker.PrintModel()
    self._textWidgetOutput.HighlightCode()

  def PrintInstantiatedModel(self):
    self._textWidgetOutput.delete('1.0', tkinter.END)
    self._walker.PrintCompiledModel()
    self._textWidgetOutput.HighlightCode()

  def StartWalk(self):
    if not self._walker.IsModelCompiled():
      numberOfErrors = self._walker.CompileModel()
      self._redirectorToTextWidgetOutput.write("Model checked: %d errors\n" % numberOfErrors)
      if not self._walker.IsModelCompiled():
        return
    ok = self._walker.StartSimulation()
    if not ok:
      return
    self._redirectorToTextWidgetOutput.write("Simulation started\n")
    states = self._walker.GetPathsOfInitialStates()
    for state in states:
      number = self._walker.NewProcess(state)
      if number!=-1:
        self._redirectorToTextWidgetOutput.write("Process %d created at %s\n" % (number, state))
    self._walker.SetBacktrackLimitHere()
    self._UpdateDisplay()

  def StopWalk(self):
    ok = self._walker.StopSimulation()
    if ok:
      self._redirectorToTextWidgetOutput.write("Simulation stopped\n")
    self._UpdateDisplay()

  def MoveCurrentProcessForward(self):
    processNumber = self._currentProcessNumber.get()
    if processNumber==0:
      return
    processBranch = self._currentProcessBranch.get()
    self._walker.MoveProcessForward(processNumber, processBranch)
    self._UpdateDisplay()

  def Backtrack(self):
    self._walker.Backtrack()
    self._UpdateDisplay()

  def FindOne(self):
    self._walker.FindOne()
    self._UpdateDisplay()

  def FindNext(self):
    self._walker.FindNext()
    self._UpdateDisplay()

  def SaveHistory(self):
    fileName = tkinter.filedialog.asksaveasfilename(filetypes = (("Scola scripts", "*.scolas"),("All files", "*.*") ))
    if fileName=='':
      return
    try:
      outputFile = open(fileName, "w")
    except:
      self.RaiseError('Unable to open file "%s"\n' % fileName)
    self._walker.SetOutputFile(outputFile)
    self._walker.PrintHistory()
    self._walker.SetOutputFile(self._redirectorToTextWidgetOutput)
    outputFile.flush()
    outputFile.close()

  def PlayHistory(self):
    fileName = tkinter.filedialog.askopenfilename(filetypes = (("Scola scripts", "*.scolas"),("All files", "*.*") ))
    if fileName=='':
      return
    try:
      inputFile = open(fileName, "r")
    except:
      self.RaiseError('Unable to open file "%s"\n' % fileName)
      return
    for line in inputFile:
      line = re.sub("#.*", "", line)
      line = line.rstrip()
      tokens = re.split("[ \t]+", line)
      if len(tokens)==0:
        continue
      if tokens[0]=="next":
        if len(tokens)<2:
          continue
        if not self.IsInteger(tokens[1]):
          continue
        processNumber = int(tokens[1])
        if len(tokens)>=3:
          processBranch = tokens[2]
        else:
          processBranch = ""
        self._walker.MoveProcessForward(processNumber, processBranch)
    inputFile.close()
    self._UpdateDisplay()

  def IsInteger(self, arg):
    try:
      int(arg)
    except:
      return False
    return True

  def SetOptions(self):
    dialog = ScolaSetOptionsDialog(self)

  @staticmethod
  def PressButton1(event):
    wizard = ScolaWizard.GetWizard()
    wizard._PressButton1(event)

  def _PressButton1(self, event):
    x = int(event.x / self.zoneSize)
    y = int(event.y / self.zoneSize)

  @staticmethod
  def MoveButton1(event):
    wizard = ScolaWizard.GetWizard()
    wizard._MoveButton1(event)

  def _MoveButton1(self, event):
    x = int(event.x / self.zoneSize)
    y = int(event.y / self.zoneSize)

  @staticmethod
  def ReleaseButton1(event):
    wizard = ScolaWizard.GetWizard()
    wizard._ReleaseButton1(event)

  def _ReleaseButton1(self, event):
    pass

  def RaiseError(self, message):
    tkinter.messagebox.showerror("error", message)
    sys.stderr.write(message)
    sys.stderr.flush()

# 4) Update
# ---------

  def _UpdateDisplay(self):
    self._UpdateDisplayOfProcesses()
    self._UpdateDisplayOfCurrentStateOfSystem()
    self._UpdateDisplayOfHistory()

  def _UpdateDisplayOfProcesses(self):
    self._textWidgetProcesses.delete('1.0', tkinter.END)
    self._walker.SetOutputFile(self._redirectorToTextWidgetProcesses)
    self._walker.PrintProcesses(self._walker.GetNumbersOfActiveProcesses())
    self._walker.SetOutputFile(self._redirectorToTextWidgetOutput)
    self._UpdateDisplayOfProcessNumbers()

  @staticmethod
  def UpdateCurrentProcessBranch(event):
    wizard = ScolaWizard.GetWizard()
    wizard._UpdateCurrentProcessBranch()

  def _UpdateDisplayOfProcessNumbers(self):
    numbers = self._walker.GetNumbersOfActiveProcesses()
    self._currentProcessNumberCombobox['values'] = numbers
    if len(numbers)==0:
      self._currentProcessNumber.set(0)
    else:
      number = self._currentProcessNumber.get()
      if not number in numbers:
        self._currentProcessNumberCombobox.current(0)
    self._UpdateCurrentProcessBranch()

  def _UpdateCurrentProcessBranch(self):
    processNumber = self._currentProcessNumber.get()
    if processNumber==0:
      self._currentProcessBranchCombobox['values'] = []
      self._currentProcessBranchCombobox.delete(0, tkinter.END)
      self._currentProcessBranch.set("")
      return
    branches = self._walker.GetBranchesOfLocationOfProcess(processNumber)
    self._currentProcessBranchCombobox['values'] = branches
    if len(branches)!=0:
      self._currentProcessBranchCombobox.current(0)
    else:
      self._currentProcessBranchCombobox.delete(0, tkinter.END)
      self._currentProcessBranch.set("")

  def _UpdateDisplayOfCurrentStateOfSystem(self):
    self._textWidgetSystem.delete('1.0', tkinter.END)
    self._walker.SetOutputFile(self._redirectorToTextWidgetSystem)
    self._walker.PrintBlocks(self._walker.GetIdentifiersOfTopLevelBlocks())
    self._textWidgetSystem.HighlightCode()
    self._walker.SetOutputFile(self._redirectorToTextWidgetOutput)

  def _UpdateDisplayOfHistory(self):
    self._textWidgetHistory.delete('1.0', tkinter.END)
    self._walker.SetOutputFile(self._redirectorToTextWidgetHistory)
    self._walker.PrintHistory()
    self._textWidgetHistory.HighlightScript()
    self._walker.SetOutputFile(self._redirectorToTextWidgetOutput)

# 5) Output redirector
# --------------------

class ScolaOutputRedirector(object):
  def __init__(self, textWidget, isErrorStream = False):
    self._textWidget = textWidget
    self._isErrorStream = isErrorStream

  def write(self, string):
    start = self._textWidget.index('insert')
    self._textWidget.insert('end', string)
    self._textWidget.see('end')
    if self._isErrorStream:
      self._textWidget.HighlightErrorMessage(start, 'end')

  def flush(self):
    pass

# 6) Dialog to set options
# ------------------------

class ScolaSetOptionsDialog(tkinter.Toplevel):
  def __init__(self, parent):
    tkinter.Toplevel.__init__(self, parent._rootWindow)
    self.transient(parent._rootWindow)
    self._parent = parent
    self.title("Walk options")
    self.wm_iconbitmap(bitmap="Wizard/AltaRicaWizard.ico")
    self._createOptionFrame()
    self.initial_focus = self._optionFrame
    self._createButtonbox()
    self.grab_set()
    if not self.initial_focus:
      self.initial_focus = self
    self.protocol("WM_DELETE_WINDOW", self.Cancel)
    self.geometry("+%d+%d" % (self._parent._rootWindow.winfo_rootx()+50, self._parent._rootWindow.winfo_rooty()+50))
    self.initial_focus.focus_set()
    self.wait_window(self)

  def _createOptionFrame(self):
    self._optionFrame = tkinter.Frame(self)
    self._optionFrame.pack(padx=5, pady=5)
    self._maximumNumberOfHistories = tkinter.IntVar()
    self._maximumSizeOfHistories = tkinter.IntVar()
    self._maximumNumberOfStepsOfWalks = tkinter.IntVar()
    self._maximumNumberOfHistoriesLabel = tkinter.Label(self._optionFrame, text="Maximum number of histories", anchor=tkinter.E)
    self._maximumSizeOfHistoriesLabel = tkinter.Label(self._optionFrame, text="Maximum size of histories", anchor=tkinter.E)
    self._maximumNumberOfStepsOfWalksLabel = tkinter.Label(self._optionFrame, text="Maximum number of steps per walk", anchor=tkinter.E)
    self._maximumNumberOfHistoriesEntry = tkinter.Entry(self._optionFrame, textvariable=self._maximumNumberOfHistories)
    self._maximumSizeOfHistoriesEntry = tkinter.Entry(self._optionFrame, textvariable=self._maximumSizeOfHistories)
    self._maximumNumberOfStepsOfWalksEntry = tkinter.Entry(self._optionFrame, textvariable=self._maximumNumberOfStepsOfWalks)
    self._maximumNumberOfHistories.set(self._parent._walker.GetMaximumNumberOfHistories())
    self._maximumSizeOfHistories.set(self._parent._walker.GetMaximumSizeOfHistories())
    self._maximumNumberOfStepsOfWalks.set(self._parent._walker.GetMaximumNumberOfStepsOfWalks())
    self._maximumNumberOfHistoriesLabel.grid(row=0, column=0, sticky=tkinter.NW+tkinter.SE)
    self._maximumNumberOfHistoriesEntry.grid(row=0, column=1, sticky=tkinter.NW+tkinter.SE)
    self._maximumSizeOfHistoriesLabel.grid(row=1, column=0, sticky=tkinter.NW+tkinter.SE)
    self._maximumSizeOfHistoriesEntry.grid(row=1, column=1, sticky=tkinter.NW+tkinter.SE)
    self._maximumNumberOfStepsOfWalksLabel.grid(row=2, column=0, sticky=tkinter.NW+tkinter.SE)
    self._maximumNumberOfStepsOfWalksEntry.grid(row=2, column=1, sticky=tkinter.NW+tkinter.SE)

  def _createButtonbox(self):
    self._buttonBox = tkinter.Frame(self)
    self._okButton = tkinter.Button(self._buttonBox, text="OK", width=10, command=self.Validate, default=tkinter.ACTIVE)
    self._okButton.pack(side=tkinter.LEFT, padx=5, pady=5)
    self._cancelButton = tkinter.Button(self._buttonBox, text="Cancel", width=10, command=self.Cancel)
    self._cancelButton.pack(side=tkinter.LEFT, padx=5, pady=5)
    self.bind("<Return>", self.Validate)
    self.bind("<Escape>", self.Cancel)
    self._buttonBox.pack()

  def Validate(self, event=None):
    self.initial_focus.focus_set() # put focus back
    self.withdraw()
    self.update_idletasks()
    self._parent._walker.SetMaximumNumberOfHistories(self._maximumNumberOfHistories.get())
    self._parent._walker.SetMaximumSizeOfHistories(self._maximumSizeOfHistories.get())
    self._parent._walker.SetMaximumNumberOfStepsOfWalks(self._maximumNumberOfStepsOfWalks.get())
    self.Cancel()

  def Cancel(self, event=None):
    # put focus back to the parent window
    self._parent._rootWindow.focus_set()
    self.destroy()
